;; Generated package description from debbugs.el  -*- no-byte-compile: t -*-
(define-package "debbugs" "0.43" "SOAP library to access debbugs servers" '((emacs "26.1") (soap-client "3.1.5")) :commit "268a377e48709b8dae88f773277215472581beaf" :url "https://elpa.gnu.org/packages/debbugs.html" :authors '(("Michael Albinus" . "michael.albinus@gmx.de")) :maintainer '("Michael Albinus" . "michael.albinus@gmx.de") :keywords '("comm" "hypermedia"))
